import torch
from torch import nn


from omegaconf.dictconfig import DictConfig


class AttentionModel(nn.Module):
    """My model"""

    def __init__(
        self,
        input_dim: int,
        num_classes: int,
        d_model: int,
        n_heads: int,
        num_layers: int,
        name: str = "model",
    ):
        """
        Model we are using with a transformer.

        Args:
            input_dim: input dimension
            num_classes: number of classes
            d_model: dimension of model
            n_heads: number of attention heads
            num_layers: number of layers
            name: model name
        """
        super().__init__()
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.d_model = d_model
        self.n_heads = n_heads
        self.num_layers = num_layers
        self.name = name

        self.input_proj = nn.Linear(input_dim, d_model)
        encoder_layer = nn.TransformerEncoderLayer(
            d_model=d_model,
            nhead=n_heads,
            batch_first=True,
        )
        self.encoder = nn.TransformerEncoder(
            encoder_layer,
            num_layers=num_layers,
        )

        self.avg_pool = nn.AdaptiveAvgPool1d(1)

        self.ff = nn.Linear(d_model, num_classes)

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "AttentionModel":
        """Create an instance of the model from config

        Args:
            cfg (DictConfig): config Object

        Raises:
            ValueError: input_dim is missing
            ValueError: num_classes is missing
            ValueError: d_model is missing
            ValueError: n_head is missing
            ValueError: num_layers is missing

        Returns:
            RModel: Instance of RModel from config
        """
        if "input_dim" not in cfg:
            raise ValueError("cfg.input_dim is required")

        if "num_classes" not in cfg:
            raise ValueError("cfg.num_classes is required")

        if "d_model" not in cfg:
            raise ValueError("cfg.d_model is required")

        if "n_head" not in cfg:
            raise ValueError("cfg.n_head is required")

        if "num_layers" not in cfg:
            raise ValueError("cfg.num_layers is required")

        if "name" not in cfg:
            name = f"model_{cfg.input_dim}"
            print(f"cfg.name wasn't defined, using default name: {name}")
        else:
            name = cfg.name

        return cls(
            cfg.input_dim,
            cfg.num_classes,
            cfg.d_model,
            cfg.n_head,
            cfg.num_layers,
            name,
        )

    def forward(
        self,
        x: [torch.Tensor, torch.Tensor],
    ) -> torch.Tensor:
        """Processes the data

        Args:
            x (torch.Tensor): data

        Returns:
            torch.Tensor: result of the processing
        """
        x, mask = x

        x = self.input_proj(x)
        x = self.encoder(x, src_key_padding_mask=~mask)

        x = x.transpose(1, 2)  # (B, d_model, T)
        x = self.avg_pool(x).squeeze(-1)  # (B, d_model)

        x = self.ff(x)

        return x
