import torch
from torch import nn


from omegaconf.dictconfig import DictConfig


class TransformerTCN(nn.Module):
    """My model"""

    def __init__(
        self,
        input_dim: int,
        num_classes: int,
        d_model: int,
        n_heads: int,
        num_layers: int,
        name: str = "model",
    ):
        """
        Model we are using with a transformer.

        Args:
            input_dim: input dimension
            num_classes: number of classes
            d_model: dimension of model
            n_heads: number of attention heads
            num_layers: number of layers
            name: model name
        """
        super().__init__()
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.d_model = d_model
        self.n_heads = n_heads
        self.num_layers = num_layers
        self.name = name

        self.conv_block = nn.Sequential(
            nn.Conv1d(input_dim, 64, kernel_size=3, dilation=1, padding=1),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Conv1d(64, 64, kernel_size=3, dilation=2, padding=2),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Conv1d(64, 64, kernel_size=3, dilation=4, padding=4),
            nn.ReLU(),
            nn.Dropout(0.3),
        )

        encoder_layer = nn.TransformerEncoderLayer(
            d_model=64,
            nhead=4,
            dim_feedforward=128,
            dropout=0.3,
            batch_first=True,
        )
        self.transformer_encoder = nn.TransformerEncoder(encoder_layer, num_layers=2)

        self.classifier = nn.Sequential(
            nn.Linear(64, 128),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(128, num_classes),
        )

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "TransformerTCN":
        """Create an instance of the model from config

        Args:
            cfg (DictConfig): config Object

        Raises:
            ValueError: input_dim is missing
            ValueError: num_classes is missing
            ValueError: d_model is missing
            ValueError: n_head is missing
            ValueError: num_layers is missing

        Returns:
            RModel: Instance of RModel from config
        """
        if "input_dim" not in cfg:
            raise ValueError("cfg.input_dim is required")

        if "num_classes" not in cfg:
            raise ValueError("cfg.num_classes is required")

        if "d_model" not in cfg:
            raise ValueError("cfg.d_model is required")

        if "n_head" not in cfg:
            raise ValueError("cfg.n_head is required")

        if "num_layers" not in cfg:
            raise ValueError("cfg.num_layers is required")

        if "name" not in cfg:
            name = f"model_{cfg.input_dim}"
            print(f"cfg.name wasn't defined, using default name: {name}")
        else:
            name = cfg.name

        return cls(
            cfg.input_dim,
            cfg.num_classes,
            cfg.d_model,
            cfg.n_head,
            cfg.num_layers,
            name,
        )

    def forward(
        self,
        x: [torch.Tensor, torch.Tensor],
    ) -> torch.Tensor:
        """Processes the data

        Args:
            x (torch.Tensor): data

        Returns:
            torch.Tensor: result of the processing
        """
        x, mask = x
        x = self.conv_block(x)  # [batch, 64, seq_length]

        x = x.permute(0, 2, 1)  # [batch, seq_length, features]

        x = self.transformer_encoder(x, src_key_padding_mask=~mask)

        # x = x.permute(0, 2, 1)  # [batch, seq_length, features]

        # Masked average pooling
        if mask is not None:
            mask = mask.unsqueeze(-1)  # [batch, seq_length, 1]
            x = (x * mask).sum(dim=1) / mask.sum(dim=1)
        else:
            x = x.mean(dim=1)

        logits = self.classifier(x)
        return logits
