from typing import Optional

from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split, KFold, StratifiedKFold

import polars as pl
import numpy as np

from scipy.spatial.transform import Rotation as R

from omegaconf.dictconfig import DictConfig


class PrepareData:
    def __init__(self):
        pass

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "PrepareData":
        return cls()

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:
        return df


class DropUnnecessaryColumns(PrepareData):

    def __init__(
        self,
        features_to_check: Optional[list[str]],
        id_features: Optional[list[str]] = None,
        target_label: str = "gesture",
    ):
        super().__init__()

        self.features_to_check = features_to_check

        self.id_features = (
            id_features
            if id_features is not None
            else ["sequence_id", "sequence_counter"]
        )

        self.target_label = target_label

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "DropUnnecessaryColumns":
        """Loads from config

        Args:
            cfg (DictConfig): config object

        Raises:
            ValueError: features_to_use missing
            ValueError: max_sequence_count missing
            ValueError: label_category_to_int missing

        Returns:
            DropUnnecessaryColumns: Instance of DropUnnecessaryColumns from config
        """
        if "features_to_check" not in cfg:
            raise ValueError("features_to_check is required")

        if "id_features" not in cfg:
            id_features = ["sequence_id", "sequence_counter"]
            print(f"id_features was not specified, using {id_features}")
        else:
            id_features = cfg.id_features

        if "target_label" not in cfg:
            target_label = "gesture"
            print(f"target_label was not specified, using: {target_label}")
        else:
            target_label = cfg.target_label

        return cls(
            features_to_check=cfg.features_to_check,
            id_features=id_features,
            target_label=target_label,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:
        features_to_keep = []
        features_to_keep += self.id_features
        features_to_keep += self.features_to_check
        if self.target_label in df:
            features_to_keep += [self.target_label]

        df = df[features_to_keep]

        return df


class AddIntegerLabel(PrepareData):

    def __init__(
        self,
        label_category_to_int: dict[str, int],
        target_label: str = "gesture",
    ):
        super().__init__()

        self.label_category_to_int = label_category_to_int

        self.target_label = target_label

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "AddIntegerLabel":
        """Loads from config

        Args:
            cfg (DictConfig): config object

        Raises:
            ValueError: features_to_use missing
            ValueError: max_sequence_count missing
            ValueError: label_category_to_int missing

        Returns:
            DropUnnecessaryColumns: Instance of DropUnnecessaryColumns from config
        """

        if "label_category_to_int" not in cfg:
            raise ValueError("label_category_to_int is required")

        if "target_label" not in cfg:
            target_label = "gesture"
            print(f"target_label was not specified, using: {target_label}")
        else:
            target_label = cfg.target_label

        return cls(
            label_category_to_int=cfg.label_category_to_int,
            target_label=target_label,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        if self.target_label in df:
            int_labels = df[[self.target_label]].map_rows(
                lambda x: self.label_category_to_int[x[0]]
            )
            df = df.with_columns(
                pl.Series(name=f"{self.target_label}_int", values=int_labels)
            )

        return df


class FilterMaxSequenceLength(PrepareData):

    def __init__(
        self,
        max_sequence_count: int = 103,
    ):
        super().__init__()
        self.max_sequence_count = max_sequence_count

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "FilterMaxSequenceLength":

        if "max_sequence_count" not in cfg:
            raise ValueError("max_sequence_count is required")

        return cls(
            max_sequence_count=cfg.max_sequence_count,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        df = df.filter(pl.col("sequence_counter") < self.max_sequence_count)

        return df


class ShrinkSequence(PrepareData):

    def __init__(
        self,
        max_sequence_count: int = 103,
    ):
        super().__init__()
        self.max_sequence_count = max_sequence_count

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "ShrinkSequence":

        if "max_sequence_count" not in cfg:
            raise ValueError("max_sequence_count is required")

        return cls(
            max_sequence_count=cfg.max_sequence_count,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        df_grouped = df.group_by("sequence_id").all()

        all_sequences = []

        for i in range(df_grouped.shape[0]):
            sequence = df_grouped[i].explode(pl.all().exclude("sequence_id"))
            if len(sequence) > self.max_sequence_count:
                take_sample = np.linspace(
                    0,
                    sequence.shape[0],
                    self.max_sequence_count,
                    endpoint=False,
                    dtype=int,
                )

                sequence = sequence[take_sample]
            all_sequences.append(sequence)

        result_df = pl.concat(all_sequences)

        return result_df


class DropNulls(PrepareData):

    def __init__(
        self,
        features_to_check: list[str],
    ):
        super().__init__()
        self.features_to_check = features_to_check

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "DropNulls":

        if "features_to_check" not in cfg:
            raise ValueError("label_category_to_int is required")

        return cls(
            features_to_check=cfg.features_to_check,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        df = df.drop_nulls(subset=self.features_to_check)

        if df.shape[0] == 0:
            raise ValueError("Some features are missing")

        return df


class FillNulls(PrepareData):

    def __init__(
        self,
        features_to_check: list[str],
        fill_value: float = 0,
    ):
        super().__init__()
        self.features_to_check = features_to_check
        self.fill_value = fill_value

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "FillNulls":

        if "features_to_check" not in cfg:
            raise ValueError("label_category_to_int is required")

        # TODO: add fill_value to config
        return cls(
            features_to_check=cfg.features_to_check,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        df = df.fill_null(self.fill_value)

        return df


class SortData(PrepareData):

    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "SortData":

        return cls()

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        df = df.sort(["sequence_id", "sequence_counter"])

        return df


class DropBadData(PrepareData):

    def __init__(
        self,
        features_to_check: list[str],
        max_sequence_count: int = 103,
    ):
        super().__init__()
        self.features_to_check = features_to_check
        self.max_sequence_count = max_sequence_count

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "DropBadData":

        if "features_to_check" not in cfg:
            raise ValueError("label_category_to_int is required")

        if "max_sequence_count" not in cfg:
            raise ValueError("max_sequence_count is required")

        return cls(
            features_to_check=cfg.features_to_check,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        df = df.filter(pl.col("sequence_counter") < self.max_sequence_count)

        df = df.drop_nulls(subset=self.features_to_check)

        if df.shape[0] == 0:
            raise ValueError("Some features are missing")

        df = df.sort(["sequence_id", "sequence_counter"])

        return df


class AddMagnitude(PrepareData):

    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "AddMagnitude":

        return cls()

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        df = df.with_columns(
            ((df["acc_x"] ** 2 + df["acc_y"] ** 2 + df["acc_z"] ** 2) ** 0.5).alias(
                "acc_magnitude"
            ),
        )

        df = df.with_columns(
            np.arccos(df["acc_x"] / df["acc_magnitude"]).alias("alpha"),  # type: ignore
        )

        df = df.with_columns(
            np.arccos(df["acc_y"] / df["acc_magnitude"]).alias("beta"),  # type: ignore
        )

        df = df.with_columns(
            np.arccos(df["acc_z"] / df["acc_magnitude"]).alias("gamma"),  # type: ignore
        )

        return df


class AddGlobalFeatures(PrepareData):

    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "AddGlobalFeatures":

        return cls()

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:

        global_acc = []

        for acc, quat in zip(
            df.select(["acc_x", "acc_y", "acc_z"]).to_numpy(),
            df.select(["rot_x", "rot_y", "rot_z", "rot_w"]).to_numpy(),
        ):
            # If I fill the null rotation values with 0
            # I want acc_global to be the same as acc.
            if (quat == 0).all():
                acc_global = acc
            else:
                acc_global = R.from_quat(quat).apply(acc)
            global_acc.append(acc_global)

        global_acc = np.array(global_acc)

        df = df.with_columns(
            [
                pl.Series(name="acc_gx", values=global_acc[:, 0]),
                pl.Series(name="acc_gy", values=global_acc[:, 1]),
                pl.Series(name="acc_gz", values=global_acc[:, 2]),
            ]
        )

        df = df.with_columns((pl.col("acc_gz") - 9.81).alias("acc_gz_no_gravity"))

        dt = 1 / 50

        vel_x = np.cumsum(df["acc_gx"].to_numpy()) * dt
        vel_y = np.cumsum(df["acc_gy"].to_numpy()) * dt
        vel_z = np.cumsum(df["acc_gz_no_gravity"].to_numpy()) * dt

        pos_x = np.cumsum(vel_x) * dt
        pos_y = np.cumsum(vel_y) * dt
        pos_z = np.cumsum(vel_z) * dt

        df = df.with_columns(
            [
                pl.Series(name="vel_x", values=vel_x),
                pl.Series(name="vel_y", values=vel_y),
                pl.Series(name="vel_z", values=vel_z),
                pl.Series(name="pos_x", values=pos_x),
                pl.Series(name="pos_y", values=pos_y),
                pl.Series(name="pos_z", values=pos_z),
            ]
        )

        return df


class NormalizeScaler(PrepareData):

    def __init__(
        self,
        features_to_use: list[str],
        scaler: StandardScaler = None,
    ):
        super().__init__()
        self.features_to_use = features_to_use
        self.scaler = scaler

    @classmethod
    def from_config(
        cls,
        cfg: DictConfig,
    ) -> "NormalizeScaler":
        if "features_to_use" not in cfg:
            raise ValueError("features_to_use is required")

        return cls(
            features_to_use=cfg.features_to_use,
        )

    def __call__(self, df: pl.DataFrame) -> pl.DataFrame:
        if self.scaler is None:
            self.scaler = StandardScaler()
            self.scaler.fit(df[self.features_to_use])

        df[list(self.features_to_use)] = self.scaler.transform(df[self.features_to_use])

        return df


def split_data(
    df: pl.DataFrame,
    valid_size: float,
    random_state: int = 42,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    """
    Split data into training and validation sets.

    Args:
        df: all polars Dataframe
        valid_size: validation size
        random_state: random state for splitting data

    Returns:
        training and validation Dataframes

    """
    df = df.group_by("sequence_id").all()
    df = df.sort("sequence_id")
    df_train, df_valid = train_test_split(
        df,
        test_size=valid_size,
        random_state=random_state,
    )

    df_train = df_train.explode(pl.all().exclude("sequence_id"))
    df_valid = df_valid.explode(pl.all().exclude("sequence_id"))

    return df_train, df_valid


def k_fold_generator(
    df: pl.DataFrame,
    number_of_folds: int,
    random_state: int = 42,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    df = df.group_by("sequence_id").all()
    df = df.sort("sequence_id")

    kf = KFold(n_splits=number_of_folds, shuffle=True, random_state=random_state)
    result = kf.split(df)

    return result


def k_fold_respect_to_label_generator(
    df: pl.DataFrame,
    number_of_folds: int,
    random_state: int = 42,
) -> tuple[pl.DataFrame, pl.DataFrame]:
    df = df.group_by("sequence_id").agg(
        pl.all().exclude("sequence_id", "gesture"),
        pl.col("gesture").first().alias("gesture"),
    )
    df = df.sort("sequence_id")

    kf = StratifiedKFold(
        n_splits=number_of_folds, shuffle=True, random_state=random_state
    )

    result = kf.split(df.to_numpy(), df["gesture"].to_numpy())

    return result
