from cmi_sensor_kaggle.utils.data_preparation import (
    DropUnnecessaryColumns,
    FillNulls,
    ShrinkSequence,
    SortData,
    AddGlobalFeatures,
    AddMagnitude,
    NormalizeScaler,
    split_data,
)
from cmi_sensor_kaggle.utils.transforms import (
    AddGaussianNoise,
    TTCNTransform,
    NormalTargetTransform,
)
from cmi_sensor_kaggle.utils.normal_data import NormalData

import torch
from torch import nn
from torch.utils.data import DataLoader

from ignite.engine import (
    Engine,
    Events,
    create_supervised_trainer,
    create_supervised_evaluator,
)
from ignite.metrics import Accuracy, Loss
from ignite.handlers import ModelCheckpoint, EarlyStopping
from ignite.handlers.tqdm_logger import ProgressBar
from ignite.contrib.handlers import TensorboardLogger, global_step_from_engine

from pathlib import Path

import polars as pl

from omegaconf import DictConfig


def get_data_loaders(cfg: DictConfig) -> tuple[DataLoader, DataLoader]:

    root_data_path = Path(
        "/Users/ramin/.cache/kagglehub/competitions/cmi-detect-behavior-with-sensor-data"
    )
    # root_data_path = Path(
    #     kagglehub.competition_download("cmi-detect-behavior-with-sensor-data")
    # )

    prs = [
        DropUnnecessaryColumns.from_config(cfg),
        FillNulls.from_config(cfg),
        ShrinkSequence.from_config(cfg),
        SortData.from_config(cfg),
        AddGlobalFeatures.from_config(cfg),
        AddMagnitude.from_config(cfg),
        NormalizeScaler.from_config(cfg),
    ]

    train_transforms = [
        AddGaussianNoise.from_config(cfg),
        TTCNTransform.from_config(cfg),
    ]

    val_transforms = [
        TTCNTransform.from_config(cfg),
    ]

    target_transform = NormalTargetTransform.from_config(cfg)

    df = pl.read_csv(root_data_path / "train.csv")

    train_df, val_df = split_data(df, 0.2, 42)

    for pr in prs:
        train_df = pr(train_df)

    for pr in prs:
        val_df = pr(val_df)

    train_data = NormalData(train_df, train_transforms, target_transform)
    val_data = NormalData(val_df, val_transforms, target_transform)

    train_loader = DataLoader(train_data, batch_size=16)
    val_loader = DataLoader(val_data, batch_size=16)

    return train_loader, val_loader


def score_function(engine):
    return engine.state.metrics["accuracy"]


def get_normal_engines_ignite(
    model: nn.Module,
    train_loader: DataLoader,
    val_loader: DataLoader,
    cfg: DictConfig,
    device: str,
    run_counter: int = 0,
) -> tuple[Engine, ModelCheckpoint]:

    if run_counter == 0:
        check_point_path = Path("checkpoints")
        if check_point_path.exists():
            p = list(Path("checkpoints").glob(f"{cfg.model.name}_*"))
            names = [x.name for x in p]
            if len(names) > 0:
                run_counter = int(sorted(names)[-1].split("_")[-1]) + 1

    print("run_counter:", run_counter)

    optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
    criterion = nn.CrossEntropyLoss()

    # Define trainer
    trainer = create_supervised_trainer(
        model=model,
        optimizer=optimizer,
        loss_fn=criterion,
        device=device,
    )

    # Define evaluators
    val_metrics = {
        "accuracy": Accuracy(),
        "loss": Loss(criterion),
    }

    train_evaluator = create_supervised_evaluator(
        model=model,
        metrics=val_metrics,
        device=device,
    )
    val_evaluator = create_supervised_evaluator(
        model=model,
        metrics=val_metrics,
        device=device,
    )

    # Add progress bar to trainer
    pbar = ProgressBar()
    pbar.attach(trainer)

    # Add progress bar to evaluators
    pbar_1 = ProgressBar()
    pbar_1.attach(train_evaluator)

    pbar_2 = ProgressBar()
    pbar_2.attach(val_evaluator)

    # Add logging of training results to trainer
    @trainer.on(Events.EPOCH_COMPLETED)
    def log_training_results(engine):
        train_evaluator.run(train_loader)
        metrics = train_evaluator.state.metrics
        print(
            f"Training Results - Epoch[{engine.state.epoch}] Avg accuracy: {metrics['accuracy']:.2f} Avg loss: {metrics['loss']:.2f}"
        )

    # Add logging of validating results to trainer
    @trainer.on(Events.EPOCH_COMPLETED)
    def log_validation_results(engine):
        val_evaluator.run(val_loader)
        metrics = val_evaluator.state.metrics
        print(
            f"Validation Results - Epoch[{engine.state.epoch}] Avg accuracy: {metrics['accuracy']:.2f} Avg loss: {metrics['loss']:.2f}"
        )

    # Add model checkpointing to val evaluator
    model_checkpoint = ModelCheckpoint(
        f"checkpoints/{cfg.model.name}_{run_counter}",
        n_saved=1,
        filename_prefix="best",
        score_function=score_function,
        score_name="accuracy",
        global_step_transform=global_step_from_engine(trainer),
    )

    val_evaluator.add_event_handler(
        Events.COMPLETED,
        model_checkpoint,
        {"model": model},
    )

    # Add Early stopping
    early_stopping = EarlyStopping(
        patience=cfg.patience,
        score_function=score_function,
        trainer=trainer,
        min_delta=cfg.min_delta,
    )

    val_evaluator.add_event_handler(
        Events.COMPLETED,
        early_stopping,
    )

    # Add tensorboard
    tb_logger = TensorboardLogger(log_dir=f"tb-logger/{cfg.model.name}_{run_counter}")

    tb_logger.attach_output_handler(
        trainer,
        event_name=Events.ITERATION_COMPLETED(every=100),
        tag="training",
        output_transform=lambda loss: {"batch_loss": loss},
    )

    for tag, evaluator in [
        ("training", train_evaluator),
        ("validation", val_evaluator),
    ]:
        tb_logger.attach_output_handler(
            evaluator,
            event_name=Events.EPOCH_COMPLETED,
            tag=tag,
            metric_names="all",
            global_step_transform=global_step_from_engine(trainer),
        )

    return trainer, model_checkpoint
