from typing import Callable

import torch
from torch.utils.data import Dataset

import polars as pl


class NormalData(Dataset):
    """Pytorch Dataset"""

    def __init__(
        self,
        df: pl.DataFrame,
        transforms: list[Callable],
        target_transform: Callable,
    ):
        """
        Make the data to a Pytorch Dataset

        Args:
            df: polars dataframe
            transforms: functions that take a sequence as polars dataframe
              and return a transformed dataframe
            target_transform: function that takes a sequence as polars
              dataframe and returns the label
        """

        self.df_grouped = df.group_by(["sequence_id"]).agg(
            pl.all().exclude("sequence_id")
        )
        self.transforms = transforms
        self.target_transform = target_transform

    def __len__(self):
        return self.df_grouped.shape[0]

    def __getitem__(self, idx) -> tuple[torch.Tensor, torch.Tensor]:
        sequence = self.df_grouped[idx]
        sequence = sequence.explode(pl.all().exclude("sequence_id"))

        label = self.target_transform(sequence)

        for transform in self.transforms:
            sequence = transform(sequence)

        return sequence, label
