from normal_data import NormalData

from typing import Callable, Optional
from .data_preparation import split_data

from pathlib import Path

import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader

import lightning as L
from torchmetrics import Accuracy, F1Score, Precision, Recall, ConfusionMatrix

import polars as pl
import numpy as np

from matplotlib import pyplot as plt
import seaborn as sns


class NormalLightningDataModule(L.LightningDataModule):
    def __init__(
        self,
        data_dir: Path,
        train_transforms: list[Callable],
        val_transforms: list[Callable],
        target_transform: Callable,
        preprocesses: list[Callable] = None,
        valid_split: float = 0.2,
    ):
        super().__init__()
        self.data_dir = data_dir
        self.train_transforms = train_transforms
        self.val_transforms = val_transforms
        self.target_transform = target_transform
        self.preprocesses = preprocesses

        self.valid_split = valid_split

        self.df = None

        self.train_data = None
        self.val_data = None

    def prepare_data(self):

        self.df = pl.read_csv(self.data_dir / "train.csv")

    def setup(self, stage: Optional[str] = None):

        if stage == "fit":
            train_df, val_df = split_data(self.df, self.valid_split)

            for preprocess in self.preprocesses:
                train_df = preprocess(train_df)

            for preprocess in self.preprocesses:
                val_df = preprocess(val_df)

            self.train_data = NormalData(
                train_df, self.train_transforms, self.target_transform
            )

            self.val_data = NormalData(
                val_df, self.val_transforms, self.target_transform
            )

    def train_dataloader(self):
        return DataLoader(
            self.train_data,
            shuffle=True,
            batch_size=16,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_data,
            shuffle=False,
            batch_size=16,
        )


class NormalLightningModel(L.LightningModule):
    """Lightning model to make training easier"""

    def __init__(self, model, criterion: nn.Module = None):
        super().__init__()
        self.model = model

        self.accuracy = Accuracy(task="multiclass", num_classes=model.num_classes)
        self.f1_score = F1Score(task="multiclass", num_classes=model.num_classes)
        self.precision = Precision(task="multiclass", num_classes=model.num_classes)
        self.recall = Recall(task="multiclass", num_classes=model.num_classes)
        self.cf = ConfusionMatrix(task="multiclass", num_classes=model.num_classes)

        if criterion is None:
            criterion = nn.CrossEntropyLoss()
        self.criterion = criterion

        self.val_predictions = []
        self.val_targets = []

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, bach_idx):
        x, label = batch
        prediction = self.model(x)
        loss = self.criterion(prediction, label)
        self.log("train_loss", loss)

        acc = self.accuracy(prediction, label)
        f1 = self.f1_score(prediction, label)
        precision = self.precision(prediction, label)
        recall = self.recall(prediction, label)

        self.log("train_accuracy", acc, on_step=True, on_epoch=True, prog_bar=True)
        self.log("train_f1_score", f1, on_step=True, on_epoch=True)
        self.log("train_precision", precision, on_step=True, on_epoch=True)
        self.log("train_recall", recall, on_step=True, on_epoch=True)

        return loss

    def validation_step(self, batch, bach_idx):
        x, label = batch
        prediction = self.model(x)
        loss = self.criterion(prediction, label)
        self.log("val_loss", loss)

        acc = self.accuracy(prediction, label)
        f1 = self.f1_score(prediction, label)
        precision = self.precision(prediction, label)
        recall = self.recall(prediction, label)

        self.log("val_accuracy", acc, on_step=False, on_epoch=True, prog_bar=True)
        self.log("val_f1_score", f1, on_step=False, on_epoch=True)
        self.log("val_precision", precision, on_step=False, on_epoch=True)
        self.log("val_recall", recall, on_step=False, on_epoch=True)

        # self.val_predictions.append(prediction)
        self.val_predictions.append(prediction.argmax(dim=1))
        self.val_targets.append(label)

        return loss

    def on_validation_epoch_end(self):
        # Concatenate all predictions and targets
        predictions = torch.cat(self.val_predictions)
        targets = torch.cat(self.val_targets)

        # Calculate confusion matrix
        cm = self.cf(predictions, targets).detach().cpu().numpy()

        num_classes = cm.shape[0]
        total = np.sum(cm)

        true_positive = np.diag(cm)
        false_positive = np.sum(cm, axis=0) - true_positive
        false_negative = np.sum(cm, axis=1) - true_positive
        true_negative = total - (true_positive + false_positive + false_negative)

        tp_df = pl.DataFrame(
            {
                "Class": [f"{i}" for i in range(num_classes)],
                "TP": true_positive,
                "FP": false_positive,
                "FN": false_negative,
                "TN": true_negative,
                "TP+FN": true_positive + false_negative,
                "TP+FP": true_positive + false_positive,
                "Precision": np.round(
                    true_positive / (true_positive + false_positive), 2
                ),
                "Recall": np.round(true_positive / (true_positive + false_negative), 2),
            }
        )

        # Plot the confusion matrix
        fig, axes = plt.subplots(1, 2, figsize=(14, 5))
        sns.heatmap(cm, annot=True, fmt="d", cmap="Blues", ax=axes[0])
        axes[0].set_xlabel("Predicted")
        axes[0].set_ylabel("True")
        axes[0].set_title("Confusion Matrix")

        axes[1].axis("off")
        axes[1].axis("tight")

        axes[1].table(
            cellText=tp_df.rows(),
            colLabels=tp_df.columns,
            loc="center",
        )

        # Add it to tensorboard
        self.logger.experiment.add_figure(
            "Confusion Matrix", fig, global_step=self.global_step
        )

        # Close the figure to prevent memory leak
        plt.close(fig)

        # Clear the lists for the new predictions and targets
        self.val_predictions.clear()
        self.val_targets.clear()

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer


class ConvLightningModel(NormalLightningModel):
    def __init__(self, model):
        super().__init__(model)

        self.example_input_array = torch.randn(1, model.input_channel, model.d_length)


class AttentionLightningModel(NormalLightningModel):
    def __init__(self, model, criterion):
        super().__init__(model, criterion)


class NormalLightningDataModuleKFold(L.LightningDataModule):
    def __init__(
        self,
        df: pl.DataFrame,
        train_transforms: list[Callable],
        val_transforms: list[Callable],
        target_transform: Callable,
        train_indexes: list[int],
        val_indexes: list[int],
        preprocesses: list[Callable] = None,
        valid_split: float = 0.2,
        number_of_folds: int = 5,
    ):
        super().__init__()
        self.df = df
        self.train_transforms = train_transforms
        self.val_transforms = val_transforms
        self.target_transform = target_transform
        self.train_indexes = train_indexes
        self.val_indexes = val_indexes
        self.preprocesses = preprocesses

        self.valid_split = valid_split
        self.number_of_folds = number_of_folds

        self.train_data = None
        self.val_data = None

    def prepare_data(self):
        pass

    def setup(self, stage: Optional[str] = None):

        if stage == "fit":

            df = self.df.group_by("sequence_id").all()
            df = df.sort("sequence_id")

            train_df = df[self.train_indexes]
            val_df = df[self.val_indexes]

            train_df = train_df.explode(pl.all().exclude("sequence_id"))
            val_df = val_df.explode(pl.all().exclude("sequence_id"))

            for preprocess in self.preprocesses:
                train_df = preprocess(train_df)

            for preprocess in self.preprocesses:
                val_df = preprocess(val_df)

            self.train_data = NormalData(
                train_df, self.train_transforms, self.target_transform
            )

            self.val_data = NormalData(
                val_df, self.val_transforms, self.target_transform
            )

    def train_dataloader(self):
        return DataLoader(
            self.train_data,
            shuffle=True,
            batch_size=16,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_data,
            shuffle=False,
            batch_size=16,
        )
