import torch


import polars as pl
import numpy as np


from omegaconf.dictconfig import DictConfig


class NormalTransform:
    def __init__(
        self,
        features_to_use: list[str],
        max_sequence_count,
    ):
        self.features_to_use = features_to_use
        self.max_sequence_count = max_sequence_count

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "NormalTransform":
        if "features_to_use" not in cfg:
            raise ValueError("features_to_use is required")

        if "max_sequence_count" not in cfg:
            raise ValueError("max_sequence_count is required")

        return cls(
            cfg.features_to_use,
            cfg.max_sequence_count,
        )

    def __call__(
        self,
        sequence: pl.DataFrame,
    ):

        num_features = len(self.features_to_use)
        input_size = self.max_sequence_count * num_features

        data_stacked = np.stack(
            [sequence[x].to_numpy() for x in self.features_to_use],
            axis=1,
        ).reshape(-1)

        data = np.zeros(input_size)
        data[: len(data_stacked)] = data_stacked
        data = data.reshape(self.max_sequence_count, num_features)
        data = torch.from_numpy(data).float()
        data = data.permute(1, 0)

        return data


class NormalTargetTransform:
    def __init__(
        self,
        target_label: str,
        label_category_to_int: dict[str, int],
    ):
        self.target_label = target_label

        self.label_category_to_int = label_category_to_int

        self.label_int_to_category = {
            value: key for key, value in label_category_to_int.items()
        }

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "NormalTargetTransform":

        if "target_label" not in cfg:
            target_label = "gesture"
            print(f"target_label was not specified, using: {target_label}")
        else:
            target_label = cfg.target_label

        if "label_category_to_int" not in cfg:
            raise ValueError("label_category_to_int is required")

        return cls(
            target_label=target_label,
            label_category_to_int=cfg.label_category_to_int,
        )

    def __call__(self, sequence: pl.DataFrame) -> torch.Tensor:
        int_label = self.label_category_to_int[sequence[self.target_label][0]]
        label = torch.tensor(int_label)
        return label


class AttentionTransform:
    def __init__(
        self,
        features_to_use: list[str],
        max_sequence_count,
    ):
        self.features_to_use = features_to_use
        self.max_sequence_count = max_sequence_count

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "AttentionTransform":
        if "features_to_use" not in cfg:
            raise ValueError("features_to_use is required")

        if "max_sequence_count" not in cfg:
            raise ValueError("max_sequence_count is required")

        return cls(
            cfg.features_to_use,
            cfg.max_sequence_count,
        )

    def __call__(
        self,
        sequence: pl.DataFrame,
    ) -> tuple[torch.Tensor, torch.Tensor]:

        num_features = len(self.features_to_use)
        input_size = self.max_sequence_count * num_features

        data_stacked = np.stack(
            [sequence[x].to_numpy() for x in self.features_to_use],
            axis=1,
        ).reshape(-1)

        data = np.zeros(input_size)
        data[: len(data_stacked)] = data_stacked
        data = data.reshape(self.max_sequence_count, num_features)
        data = torch.from_numpy(data).float()

        attention_mask = np.zeros(self.max_sequence_count, dtype=bool)
        valid_length = len(data_stacked) // num_features
        attention_mask[:valid_length] = True
        attention_mask = torch.from_numpy(attention_mask).bool()

        return data, attention_mask


class TTCNTransform:
    def __init__(
        self,
        features_to_use: list[str],
        max_sequence_count,
    ):
        self.features_to_use = features_to_use
        self.max_sequence_count = max_sequence_count

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "TTCNTransform":
        if "features_to_use" not in cfg:
            raise ValueError("features_to_use is required")

        if "max_sequence_count" not in cfg:
            raise ValueError("max_sequence_count is required")

        return cls(
            cfg.features_to_use,
            cfg.max_sequence_count,
        )

    def __call__(
        self,
        sequence: pl.DataFrame,
    ) -> tuple[torch.Tensor, torch.Tensor]:

        num_features = len(self.features_to_use)
        input_size = self.max_sequence_count * num_features

        data_stacked = np.stack(
            [sequence[x].to_numpy() for x in self.features_to_use],
            axis=1,
        ).reshape(-1)

        data = np.zeros(input_size)
        data[: len(data_stacked)] = data_stacked
        data = data.reshape(self.max_sequence_count, num_features)
        data = torch.from_numpy(data).float()
        data = data.permute(1, 0)

        attention_mask = np.zeros(self.max_sequence_count, dtype=bool)
        valid_length = len(data_stacked) // num_features
        attention_mask[:valid_length] = True
        attention_mask = torch.from_numpy(attention_mask).bool()

        return data, attention_mask


class AddGaussianNoise:
    def __init__(
        self,
        features_to_use: list[str],
        probability: float = 0.5,
        noise_std: float = 0.01,
    ):
        self.features_to_use = features_to_use
        self.probability = probability
        self.noise_std = noise_std

    @classmethod
    def from_config(cls, cfg: DictConfig) -> "AddGaussianNoise":
        if "features_to_use" not in cfg:
            raise ValueError("features_to_use is required")

        if "gaussian_noise" in cfg and "probability" in cfg.gaussian_noise:
            probability = cfg.gaussian_noise.probability
        else:
            probability = 0.5
            print(f"Using default probability: {probability}")

        if "gaussian_noise" in cfg and "noise_std" in cfg.gaussian_noise:
            noise_std = cfg.gaussian_noise.noise_std
        else:
            noise_std = 0.01
            print(f"Using default gaussian noise_std: {noise_std}")

        return cls(
            cfg.features_to_use,
            probability=probability,
            noise_std=noise_std,
        )

    def __call__(
        self,
        sequence: pl.DataFrame,
    ):

        result = sequence

        if np.random.rand() < self.probability:
            for feature in self.features_to_use:
                result = result.with_columns(
                    pl.col(feature)
                    + np.random.normal(
                        loc=0.0, scale=self.noise_std, size=result[feature].shape
                    ),
                )

        return result
